from typing import Any, Optional, Dict
from sttn.network import SpatioTemporalNetwork
from jinja2 import Environment, PackageLoader, Template

from sttn.nli import Query
from sttn.data.brno import HealthcareDataProvider
from sttn.data.lehd import OriginDestinationEmploymentDataProvider
from sttn.data.nyc import NycTaxiDataProvider, Service311RequestsDataProvider

import pandas as pd

DATA_PROVIDERS = [NycTaxiDataProvider, Service311RequestsDataProvider, OriginDestinationEmploymentDataProvider, HealthcareDataProvider]


class Context:
    def __init__(self, query: Query):
        self._query: Query = query

        self._data_provider_id: Optional[str] = None  # data provider id
        self._data_provider_cls = None
        self._data_provider_instance = None

        self._data_provider_args: Optional[Dict[str, str]] = None

        self._analysis_code: Optional[str] = None

        self._network: Optional[SpatioTemporalNetwork] = None

        self._result: Optional[Any] = None

        import sttn.algorithms.community.detection
        import sttn.plot

    @property
    def query(self):
        return self._query

    @property
    def data_provider(self):
        return self._data_provider_instance

    @data_provider.setter
    def data_provider(self, data_provider_id: str):
        self._data_provider_id = data_provider_id
        self._data_provider_cls = Context._get_data_provider_by_id(data_provider_id)
        self._data_provider_instance = self._data_provider_cls()

    @property
    def network(self):
        return self._network

    @network.setter
    def network(self, network: SpatioTemporalNetwork):
        self._network = network

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, result: Any):
        self._result = result

    @property
    def data_provider_id(self):
        return self._data_provider_id
    
    @data_provider_id.setter
    def data_provider_id(self, data_provider_id: str):
        self._data_provider_id = data_provider_id
        
    @property
    def data_provider_args(self):
        return self._data_provider_args

    @data_provider_args.setter
    def data_provider_args(self, _data_provider_args: Dict[str, str]):
        self._data_provider_args = _data_provider_args

    @property
    def analysis_code(self):
        return self._analysis_code

    @analysis_code.setter
    def analysis_code(self, analysis_code: str):
        self._analysis_code = analysis_code

    @staticmethod
    def _get_data_provider_by_id(data_provider_id: str):
        for data_provider in DATA_PROVIDERS:
            if data_provider.__name__ == data_provider_id:
                return data_provider

        raise KeyError(f"Can not find data provider with {data_provider_id} id.")


class PromptGenerator:

    @staticmethod
    def get_df_description(df: pd.DataFrame) -> str:
        schema_str = ""
        for col, dtype in df.dtypes.items():
            if dtype.name == 'geometry':
                schema_str += f"{col}: geometry\n"
            elif pd.api.types.is_numeric_dtype(dtype):
                # Calculate min, max, and mean for numeric columns
                min_val = df[col].min()
                max_val = df[col].max()
                mean_val = df[col].mean()
                schema_str += f"{col:20}: {dtype} - Min: {min_val}, Max: {max_val}, Avg: {mean_val:.2f}\n"
            else:
                # Find the 5 most common values for string columns
                common_vals = df[col].value_counts().head(5)
                common_vals_str = ", ".join([f"{val} ({count})" for val, count in common_vals.items()])
                schema_str += f"{col:20}: {dtype} - Most common values: {common_vals_str}\n"
        return schema_str

    @staticmethod
    def get_template(template_fname: str) -> Template:
        environment = Environment(loader=PackageLoader("sttn", package_path="nli/templates"))
        return environment.get_template(template_fname)

    @staticmethod
    def generate_provider_prompt(query: Query) -> str:
        template = PromptGenerator.get_template("data_provider.j2")
        jcontext = {
            "user_query": query.query,
            "data_providers": DATA_PROVIDERS,
        }

        prompt_str = template.render(jcontext)
        return prompt_str

    @staticmethod
    def generate_data_retrieval_prompt(context: Context) -> str:
        template = PromptGenerator.get_template("data_provider_arguments.j2")

        data_provider = context.data_provider.__class__
        data_provider_instance = context.data_provider

        jcontext = {
            "user_query": context.query.query,
            "data_provider_documentation": data_provider.__doc__,
            "data_description": data_provider_instance.get_data.__doc__,
        }

        prompt_str = template.render(jcontext)
        return prompt_str

    @staticmethod
    def generate_data_filtering_prompt(context: Context) -> str:
        template = PromptGenerator.get_template("data_filter.j2")

        data_provider = context.data_provider.__class__
        data_provider_instance = context.data_provider
        nodes_description = PromptGenerator.get_df_description(context.network.nodes)
        edges_description = PromptGenerator.get_df_description(context.network.edges)

        jcontext = {
            "user_query": context.query.query,
            "data_provider_documentation": data_provider.__doc__,
            "data_provider_arguments": context.data_provider_args,
            "data_description": data_provider_instance.get_data.__doc__,
            "nodes_description": nodes_description,
            "edges_description": edges_description
        }

        prompt_str = template.render(jcontext)
        return prompt_str

    @staticmethod
    def generate_data_analysis_prompt(context: Context) -> str:
        template = PromptGenerator.get_template("data_provider_arguments.j2")

        data_provider = context.data_provider.__class__
        data_provider_instance = context.data_provider

        environment = Environment(loader=PackageLoader("sttn", package_path="nli/templates"))
        template = environment.get_template("data_analysis.j2")
        jcontext = {
            "user_query": context.query.query,
            "data_provider_documentation": data_provider.__doc__,
            "data_provider_arguments": context.data_provider_args,
            "data_description": data_provider_instance.get_data.__doc__,
        }

        prompt_str = template.render(jcontext)
        return prompt_str

    @staticmethod
    def fix_analysis_code_prompt(context: Context, exc_str: str) -> str:
        environment = Environment(loader=PackageLoader("sttn", package_path="nli/templates"))
        template = environment.get_template("code_execution_error.j2")
        jcontext = {
            "user_query": context.query.query,
            "executed_code": context.analysis_code,
            "error": exc_str,
        }

        prompt_str = template.render(jcontext)
        return prompt_str



import os
from datetime import datetime

import geopandas as gpd
import pandas as pd
import pyarrow as pa
import pyarrow.parquet
import numpy as np
from dateutil.relativedelta import relativedelta

from sttn import network
from .data_provider import DataProvider

TAXI_ZONE_SHAPE_URL = 'https://data.cityofnewyork.us/api/geospatial/d3c5-ddgc?method=export&format=Shapefile'


class NycTaxiDataProvider(DataProvider):
    """New York Taxi data provider, builds a network where nodes represent taxi zones and edges
    represent taxi trips for a given month. Yellow and green taxi trip records include fields capturing
    pick-up and drop-off dates/times, pick-up and drop-off locations, trip distances, itemized fares,
    rate types, payment types, and driver-reported passenger counts."""

    @staticmethod
    def build_network(taxi_trips, taxi_zones) -> network.SpatioTemporalNetwork:
        edges = taxi_trips.rename(
            columns={'PULocationID': 'origin', 'DOLocationID': 'destination', 'tpep_pickup_datetime': 'time'})
        edges_casted = edges.astype({'origin': 'int64', 'destination': 'int64'})
        taxi_zones = taxi_zones.rename(columns={'objectid': 'id'}).astype({'id': 'int32'})
        taxi_zones = taxi_zones.set_index('id')
        return network.SpatioTemporalNetwork(nodes=taxi_zones, edges=edges_casted)

    def get_data(self, taxi_type: str, month: str) -> network.SpatioTemporalNetwork:
        """
        Retrieves New York City taxi data

        Args:
            taxi_type (str): String taxi type one of the following values:
                'yellow' - Yellow taxi
                'green' - Green taxi
                'fhv' - For-Hire vehicles
                'fhvhv' - High-volume for-hire vehicles
            month (str): A string with year and month in the "YYYY-MM" format.
                The earliest dataset is available for 2009.

        Returns:
            SpatioTemporalNetwork: An STTN network where node represent New York City taxi zones
                and edges represent individual trips.

            The nodes dataframe contains the following columns:
                'id' (int64) - index column, represents taxi zone id
                'borough' (str) - taxi zone borough
                'zone' (str) - taxi zone name
                'geometry' (shape) - shape object for the zone

            The edges dataframe contains the following columns:
                'origin' (int64) - trip origin taxi zone id
                'destination' (int64) - trip destination taxi zone id
                'time' (datetime64[ns]) - trip start time
                'passenger_count' (int64) - number of passengers
                'fare_amount' (float64) - trip fare in USD
        """
        url = f'https://d37ci6vzurychx.cloudfront.net/trip-data/{taxi_type}_tripdata_{month}.parquet'
        taxi_data = self.cache_file(url)
        column_names = ['PULocationID', 'DOLocationID', 'tpep_pickup_datetime', 'passenger_count', 'fare_amount']
        df = pd.read_parquet(taxi_data, columns=column_names)
        df = df[(df['PULocationID'] > 0) & (df['PULocationID'] < 264)]
        df = df[(df['DOLocationID'] > 0) & (df['DOLocationID'] < 264)]
        df = df.dropna()

        from_date = datetime.strptime(month, '%Y-%m')
        to_date = from_date + relativedelta(months=1)
        df = df[(df['tpep_pickup_datetime'] >= from_date) & (df['tpep_pickup_datetime'] <= to_date)]
        df['passenger_count'] = df['passenger_count'].astype(int)
        labels = gpd.read_file(TAXI_ZONE_SHAPE_URL)
        return self.build_network(df, labels)


class Service311RequestsDataProvider(DataProvider):
    """New York City 311 request data provider, builds a network where nodes represent zip codes
     and every edge represents a 311 incident where origin and destination point to the node
     where the incident happened."""

    def build_network(self, requests, nyc_zip_shape):
        requests['from'] = requests['Incident Zip']
        requests['to'] = requests['Incident Zip']
        column_map = {'Latitude': 'latitude', 'Longitude': 'longitude', 'Complaint Type': 'complaint_type',
                      'Created Date': 'time', 'City': 'city'}
        requests = requests.rename(columns=column_map)
        requests = requests.drop('Incident Zip', 1)

        node_labels = nyc_zip_shape[['ZIPCODE', 'COUNTY', 'PO_NAME', 'geometry']]
        node_labels.columns = node_labels.columns.str.lower()
        node_labels = node_labels.rename(columns={'zipcode': 'id'})
        node_labels = node_labels.set_index('id')
        return network.SpatioTemporalNetwork(nodes=node_labels, edges=requests)

    def get_data(self, from_date, to_date):
        data = self.cache_file('https://data.cityofnewyork.us/api/views/erm2-nwe9/rows.csv')
        column_names = ['Incident Zip', 'City', 'Latitude', 'Longitude', 'Complaint Type', 'Created Date']
        filtered_file = self.filter_requests(data, from_date, to_date, column_names)

        nyc_shape = gpd.read_file(
            'https://data.cityofnewyork.us/api/views/i8iw-xf4u/files/YObIR0MbpUVA0EpQzZSq5x55FzKGM2ejSeahdvjqR20?filename=ZIP_CODE_040114.zip')
        nyc_shape['ZIPCODE'] = nyc_shape['ZIPCODE'].astype(int)
        requests = pd.read_parquet(filtered_file)
        requests['Incident Zip'] = requests['Incident Zip'].astype(int)

        return self.build_network(requests, nyc_shape)

    def filter_requests(self, requests_file, from_date, to_date, column_names):
        arg_hash = self.hash_args(from_date=from_date.timestamp(), to_date=to_date.timestamp(),
                                  column_names=column_names)
        local_filename = arg_hash + '.parquet'
        file_path = os.path.join(self.cache_dir(), local_filename)

        if not os.path.exists(file_path):
            pqwriter = None
            # write to a temporary file in order to avoid incomplete results
            tmp_file_name = arg_hash + '_tmp.parquet'
            tmp_file_path = os.path.join(self.cache_dir(), tmp_file_name)
            if os.path.exists(tmp_file_path):
                os.remove(tmp_file_path)
            types = {'Incident Zip': pd.StringDtype(), 'Complaint Type': pd.StringDtype(), 'City': pd.StringDtype()}
            chunks_iter = pd.read_csv(requests_file, parse_dates=['Created Date'], usecols=column_names, dtype=types,
                                      index_col=False, iterator=True, chunksize=1 << 18)
            for chunk in chunks_iter:
                filtered_chunk = chunk[(chunk['Created Date'] >= from_date) & (chunk['Created Date'] <= to_date)]
                filtered_chunk = filtered_chunk.dropna(subset=['Incident Zip'])
                table = pa.Table.from_pandas(df=filtered_chunk)
                if not pqwriter:
                    pqwriter = pa.parquet.ParquetWriter(tmp_file_path, table.schema)
                pqwriter.write_table(table)

            # close the parquet writer
            if pqwriter:
                pqwriter.close()
            os.rename(tmp_file_path, file_path)
            
        return file_path


class RestaurantInspectionDataProvider(DataProvider):
    """New York City Restaurant Inspections data provider builds a bipartite network where the 
     first type of nodes represents restaurants (establishments), the second type of nodes 
     represents types of inspection, and every edge represents an inspection details where the
     origin is the inspection type, and destination is the restaurant where the inspection happened."""
  
    @staticmethod
    def get_data(self, from_date: str, to_date: str) -> network.SpatioTemporalNetwork:
        """
        Retrieves New York City Restaurant Inspections data

        Args:
            from_date (str): A string with year and month and in the "YYYY-MM-DD" format.
            Specifies the start date to get data from
            
            to_date (str): A string with year, month, and day in the "YYYY-MM-DD" format.
            Specifies the end date to get data from
                

        Returns:
            SpatioTemporalNetwork: An STTN network where the first type of nodes represents restaurants (establishments),
                second type of nodes represents types of inspection, and edges represent inspections.
            
            The nodes dataframe contains the following columns:
                'ID' (str) - represents either restaurant id, also referred as 'camis', or an inspection type which 
                             is a combination of the inspection program and the type of inspection performed
                'IS_RESTAURANT' (bool) - boolean column representing whether the specific row is 
                                         referring to a restaurant (True) or an inspection type (False) 
                'NAME' (str) - restaurant name, also referred as DBA ('doing business as')
                'BORO' (str) - restaurant location borough
                'BUILDING' (str) - restaurant location building code
                'STREET' (str) - restaurant location street name
                'ZIPCODE' (float32) - restaurant location zip code
                'PHONE' (float32) - restaurant phone number
                'CUISINE_DESCRIPTION' (str) - restaurant cuisine description
                'COUNCIL_DISTRICT' (float32) - restaurant council district number
                'CENSUS_TRACT' (float32) - restaurant census tract number
                'COMMUNITY_BOARD' (float32) - restaurant community board number
                'BIN' (float32) - restaurant Building Identification Number
                'BBL' (float32) - restaurant Borough-Block-and-Lot number
                'NTA' (str) - restaurant Neighborhood Tabulation Area code
                'geometry' (geometry) - geometry object for the restaurant location
                

            The edges dataframe contains the following columns:
                'ORIGIN' (str) - the inspection type executed
                'DESTINATION' (str) - inspected restaurant id
                'INSPECTION_DATE' (datetime64[ns]) - inspection date
                'ACTION' (str) - action associated with each restaurant inspection
                'VIOLATION_CODE' (str) - violation code associated with a restaurant inspection
                'VIOLATION_DESCRIPTION' (str) - violation description associated with an
                                                establishment (restaurant) inspection
                'CRITICAL_FLAG' (str) - indicator of critical violation
                'SCORE' (float32) - total score for a particular inspection
                'GRADE' (str) - grade associated with the inspection
                'GRADE_DATE' (datetime64[ns]) - the date when the grade was issued to the establishment (restaurant)
                'RECORD_DATE' (datetime64[ns]) - the date when the record was added to the dataset
       
        """
        url = 'https://data.cityofnewyork.us/api/views/43nn-pn8j/rows.csv?accessType=DOWNLOAD'
        rest_ins_data = self.cache_file(url)
        df = pd.read_csv(rest_ins_data,
                        parse_dates=['INSPECTION DATE', 'RECORD DATE', 'GRADE DATE'],
                        infer_datetime_format=True)

        df = format_data(df, from_date, to_date)

        return build_network(df)


    def format_data(self, df: pd.DataFrame, from_date: str, to_date: str):
        """Format and clean the data."""

        # Drop values with unknown latitude and longitude
        df = df.loc[~((df['Latitude'].isna()) | (df['Longitude'].isna())
                |(df['Latitude'].round() == 0) | (df['Longitude'].round() == 0))]
        df = df.drop(columns=['Location Point1'])

        # Drop values where INSPECTION DATE == 01.01.1900 (which means they didn't have an inspection yet)
        df = df.loc[df['INSPECTION DATE'] != '01.01.1900']
        
        # Filter by INSPECTION DATE
        df = df.loc[(df['INSPECTION DATE'] >= from_date) & (df['INSPECTION DATE'] <= to_date)]
        
        # Change 0's and empty phone numbers to NaNs in BORO and PHONE
        df.loc[:, 'BORO'] = df['BORO'].map({0:np.nan})
        df.loc[:, 'PHONE'] = df['PHONE'].map({'__________': np.nan}).astype(np.float32)
        
        # Uppercase column names and replace spaces with underscore
        df.columns = df.columns.str.upper().str.strip().str.replace(' ', '_')
        
        # Change the names of some columns to more straightforward ones
        df = df.rename(columns={'CAMIS': 'ID',#'id',
                                'DBA': 'NAME'})
        return df


    def build_network(self, df: pd.DataFrame) -> network.SpatioTemporalNetwork:
        """Build an STTN network from filtered data."""

        # Separate features for edges and two types of nodes
        rest_node_features = ['ID', 'NAME', 'BORO', 'BUILDING', 'STREET', 'ZIPCODE', 'PHONE',
            'CUISINE_DESCRIPTION', 'LATITUDE', 'LONGITUDE', 'COMMUNITY_BOARD',
            'COUNCIL_DISTRICT', 'CENSUS_TRACT', 'BIN', 'BBL', 'NTA']
        insp_node_features = ['INSPECTION_TYPE']
        edge_features = ['ID', 'INSPECTION_TYPE','INSPECTION_DATE', 'ACTION', 'VIOLATION_CODE', 'VIOLATION_DESCRIPTION',
            'CRITICAL_FLAG', 'SCORE', 'GRADE', 'GRADE_DATE', 'RECORD_DATE',
            ]

        # Geodf with restaurants
        gdf_rest_nodes = df.loc[:, rest_node_features]
        gdf_rest_nodes.loc[:, 'IS_REST'] = True
        gdf_rest_nodes = gpd.GeoDataFrame(
          gdf_rest_nodes.drop(columns=['LATITUDE','LONGITUDE']),
          geometry=gpd.points_from_xy(x=gdf_rest_nodes['LATITUDE'], y=gdf_rest_nodes['LONGITUDE'])
          )
        gdf_rest_nodes.loc[:, 'ID'] = gdf_rest_nodes['ID'].astype(str)
        
        # df with inspection types
        df_insp_nodes = df.loc[:, insp_node_features]
        df_insp_nodes = df_insp_nodes.rename(columns={'INSPECTION_TYPE':'ID'})

        # Geodf with all nodes
        gdf_nodes = pd.concat([gdf_rest_nodes, df_insp_nodes], join='outer', ignore_index=True)

        # Fill NaNs in IS_REST with False
        gdf_nodes.loc[:, 'IS_REST'] = gdf_nodes['IS_REST'].fillna(False)

        # Set ID as index
        gdf_nodes = gdf_nodes.set_index('ID')

        # df with edges
        df_edges = df.loc[:, edge_features]
        df_edges.loc[:, 'ORIGIN'] = df_edges['INSPECTION_TYPE']
        df_edges.loc[:, 'DESTINATION'] = df_edges['ID'].astype(str)
        df_edges = df_edges.drop(columns=['ID', 'INSPECTION_TYPE'])

        # Downcast float64 to float32 to reduce memory footprint
        for col in gdf_nodes.select_dtypes(include=['float64']).columns:
              gdf_nodes[col] = gdf_nodes[col].astype('float32')
        for col in df_edges.select_dtypes(include=['float64']).columns:
              df_edges[col] = df_edges[col].astype('float32')

        return network.SpatioTemporalNetwork(nodes=gdf_nodes,
                                            edges=df_edges,
                                            origin='ORIGIN',
                                            destination='DESTINATION',
                                            node_id='ID')



from typing import Dict

from langchain_core.pydantic_v1 import BaseModel, Field


class DataProviderModel(BaseModel):
    provider_id: str = Field(
        description="the provider id that can be used to answer the question above,"
                    " or an empty string if none of the providers from the list can be used.")
    justification: str = Field(description="reasoning for the provider selection")


class DataProviderArgumentsModel(BaseModel):
    feasible: bool = Field(
        description="True` if the data provider can retrieve data for the user query and `False` otherwise")
    justification: str = Field(description="feasibility justification")
    arguments: Dict[str, str] = Field(
        "Dictionary with data provider argument names as keys and argument values as values")


class DataAnalysisModel(BaseModel):
    analysis_code: str = Field(description="Python code that returns the answer to the user query")



from langchain.chains import LLMChain
from langchain.output_parsers import PydanticOutputParser

from sttn.nli.prompts import Context
from sttn.nli.prompts import PromptGenerator
from sttn.nli.models.output import DataProviderModel, DataProviderArgumentsModel


class NetworkBuilder:
    def __init__(self, model: LLMChain):
        self.model = model

    def pick_data_provider(self, context: Context) -> DataProviderModel:
        parser = PydanticOutputParser(pydantic_object=DataProviderModel)
        prompt = PromptGenerator.generate_provider_prompt(context.query)
        output = self.model.predict(human_input=prompt)
        return parser.parse(self._sanitize_json_output(output))

    def pick_provider_arguments(self, context: Context) -> DataProviderArgumentsModel:
        parser = PydanticOutputParser(pydantic_object=DataProviderArgumentsModel)
        prompt = PromptGenerator.generate_data_retrieval_prompt(context)
        output = self.model.predict(human_input=prompt)
        return parser.parse(self._sanitize_json_output(output))

    def get_filtering_code(self, context: Context) -> str:
        prompt = PromptGenerator.generate_data_filtering_prompt(context)
        output = self.model.predict(human_input=prompt)
        return output

    def get_analysis_code(self, context: Context) -> str:
        prompt = PromptGenerator.generate_data_analysis_prompt(context)
        output = self.model.predict(human_input=prompt)
        return self._sanitize_output(output)

    def get_fixed_code(self, context: Context, exc: Exception) -> str:
        exc_str = self._describe_exc(exc=exc)
        prompt = PromptGenerator.fix_analysis_code_prompt(context=context, exc_str=exc_str)
        output = self.model.predict(human_input=prompt)
        return self._sanitize_output(output)

    @staticmethod
    def _sanitize_output(text: str) -> str:
        _, after = text.split("```python")
        return after.split("```")[0]

    @staticmethod
    def _sanitize_json_output(text: str) -> str:
        _, after = text.split("```json")
        return after.split("```")[0]

    @staticmethod
    def _describe_exc(exc: Exception) -> str:
        error_type = type(exc).__name__
        error_message = str(exc)
        readable_error_string = f"{error_type}: {error_message}"
        return readable_error_string



from typing import Optional

from langchain.chains import LLMChain
from langchain.memory import ConversationBufferMemory
from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    MessagesPlaceholder,
)
from langchain_core.messages import SystemMessage
from langchain_openai import ChatOpenAI

from sttn.nli import Query
from sttn.nli.data import NetworkBuilder
from sttn.nli.prompts import Context


class STTNAnalyst:
    def __init__(self, verbose: bool = False, model_name: str = "gpt-4o"):
        self._verbose = verbose
        self._model = ChatOpenAI(temperature=0, model_name=model_name)
        prompt = ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content="You are a chatbot having a conversation with a human."
                ),  # The persistent system prompt
                MessagesPlaceholder(
                    variable_name="chat_history"
                ),  # Where the memory will be stored.
                HumanMessagePromptTemplate.from_template(
                    "{human_input}"
                ),  # Where the human input will injected
            ]
        )

        memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
        self._chain = LLMChain(
            llm=self._model,
            prompt=prompt,
            verbose=verbose,
            memory=memory,
        )
        self._network_builder = NetworkBuilder(model=self._chain)
        self._context: Optional[Context] = None
        self._model_name: str = model_name

    def clarify(self, human_input: str) -> str:
        return self._chain.predict(human_input=human_input)

    def chat(self, user_query: Optional[str] = None) -> Context:
        if user_query is None:
            print("Enter your question please:")
            user_query = input()
        query = Query(user_query)
        context = Context(query=query)
        self._context = context
        data_provider = self._network_builder.pick_data_provider(context=context)
        provider_id = data_provider.provider_id
        provider_descr = data_provider.justification

        if len(provider_id) == 0:
            print(f"Don't have the data to answer the query, {provider_descr}.")
            return context

        if self._verbose:
            print(f"Picked data provider {provider_id}")
            print(provider_descr)

        context.data_provider = provider_id
        data_provider_args = self._network_builder.pick_provider_arguments(context)
        args_descr = data_provider_args.justification
        context.data_provider_args = data_provider_args.arguments

        if not data_provider_args.feasible:
            print(f"Can not retrieve the data {args_descr}.")
            return context

        if self._verbose:
            print(args_descr)
            print(f"Data provider arguments: {data_provider_args.arguments}")

        data_provider = context.data_provider

        if self._verbose:
            print(f"Retrieving the data using {provider_id} provider with the following arguments {data_provider_args}")
        network = data_provider.get_data(**data_provider_args.arguments)
        context.network = network

        filtering_code = self._network_builder.get_filtering_code(context=context)
        # generated filtering predicates are used only as a chain-of-thought at this moment
        print(filtering_code)

        analysis_code = self._network_builder.get_analysis_code(context=context)
        content = analysis_code
        prefix = "sttn_network = analyst._context.network\n"
        analysis_code = prefix + content
        context.analysis_code = content

        get_ipython().set_next_input(analysis_code)
        result = get_ipython().run_cell(analysis_code)

        if result.error_in_exec is not None:
            fixed_code = self._network_builder.get_fixed_code(context=context, exc=result.error_in_exec)
            get_ipython().set_next_input(fixed_code)
            print(fixed_code)
        else:
            context.result = result.result

        return context



import pandas as pd

from sttn import network
from . import census
from .data_provider import DataProvider


class OriginDestinationEmploymentDataProvider(DataProvider):
    """Longitudinal Employer-Household Dynamics Origin-Destination Employment Statistics
    data provider. Every node represents a census tract and every edge contains the number
    of people living in the origin and commuting to the destination tract.

    Data spec: https://lehd.ces.census.gov/data/lodes/LODES7/LODESTechDoc7.5.pdf
    """

    def build_network(self, state: str, year: int) -> network.SpatioTemporalNetwork:
        od_columns = ['w_geocode', 'h_geocode', 'S000', 'SA01', 'SA02', 'SA03', 'SE01', 'SE02', 'SE03', 'SI01', 'SI02',
                      'SI03']
        xwalk_columns = ['tabblk2010', 'trct', 'ctyname', 'zcta']
        od_data = pd.read_csv(self.od_fname, compression='gzip', usecols=od_columns)
        xwalk_data = pd.read_csv(self.xwalk_fname, compression='gzip', index_col='tabblk2010', usecols=xwalk_columns)
        # map census Block Codes to Census Tract codes
        home_joined = xwalk_data[['trct']].merge(od_data, left_index=True, right_on='h_geocode', how='inner').rename(
            columns={'trct': 'origin'})
        joined = xwalk_data[['trct']].merge(home_joined, left_index=True, right_on='w_geocode', how='inner').rename(
            columns={'trct': 'destination'})
        cleaned = joined.drop(['w_geocode', 'h_geocode'], axis=1)
        aggregated_edges = cleaned.groupby(['origin', 'destination']).sum().reset_index()

        rename_map = {'trct': 'id', 'ctyname': 'county', 'zcta': 'zip'}
        renamed = xwalk_data.reset_index()[['trct', 'ctyname', 'zcta']].rename(columns=rename_map)
        # 99999 is used for unknown zip codes
        tract_to_zip = renamed[renamed.zip != 99999].groupby('id').first()
        tract_to_zip['zip'] = tract_to_zip['zip'].astype(str)
        tract_geo_columns = ['GEOID', 'geometry']
        tract_shapes = census.get_tract_geo(state=state, year=year)
        # filter out water-only tracts:
        filtered_tracts = tract_shapes[tract_shapes.ALAND > 0]
        indexed_tracts = filtered_tracts[tract_geo_columns].set_index('GEOID')
        tracts_with_zip = indexed_tracts.merge(tract_to_zip, left_index=True, right_on='id', how='inner')

        # filter out edges for filtered nodes
        ids_to_keep = tracts_with_zip.index
        filtered_edges = aggregated_edges[
            aggregated_edges.origin.isin(ids_to_keep) & aggregated_edges.destination.isin(ids_to_keep)]
        return network.SpatioTemporalNetwork(nodes=tracts_with_zip, edges=filtered_edges)


    def get_data(self, state: str, year: int, part: str = 'main', job_type: int = 0) -> network.SpatioTemporalNetwork:
        """
        Retrieves LEHD Origin-Destination Employment Statistics for a given state and year

        Args:
            state (str): lowercase, 2-letter postal code for a chosen state
            year (str): Year of job data, starting from 2002

        Returns:
            SpatioTemporalNetwork: An STTN network where node represent census tracts and edges represent employment
                statistics for people who live in the origin tract and work in the destination tract area.

            The nodes dataframe contains the following columns:
                'id' (int64) - index column, represents census tract ids
                'county' (str) - county of the tract (e.g. "Queens County, NY")
                'zip' (int) - zip code of the tract
                'geometry' (shape) - shape object for the tract

            The edges dataframe contains the following columns:
                'origin' (int64) - origin Census tract id
                'destination' (int64) - destination Census tract id
                'S000' (int32) - Num Total number of jobs
                'SA01' (int32) - Num Number of jobs of workers age 29 or younger
                'SA02' (int32) - Num Number of jobs for workers age 30 to 54
                'SA03' (int32) - Num Number of jobs for workers age 55 or older
                'SE01' (int32) - Num Number of jobs with earnings $1250/month or less
                'SE02' (int32) - Num Number of jobs with earnings $1251/month to $3333/month
                'SE03' (int32) - Num Number of jobs with earnings greater than $3333/month
                'SI01' (int32) - Num Number of jobs in Goods Producing industry sectors
                'SI02' (int32) - Num Number of jobs in Trade, Transportation, and Utilities industry sectors
                'SI03' (int32) - Num Number of jobs in All Other Services industry sectors
        """
        self._cache(state=state.lower(), year=year, part=part, job_type=job_type)
        return self.build_network(state=state, year=year)

    def _cache(self, state: str, year: int, part: str = 'main', job_type: int = 0) -> None:
        od_fname = '{state}_od_{part}_JT0{job_type}_{year}.csv.gz'.format(
            state=state, part=part, job_type=job_type, year=year)
        xwalk_fname = '{state}_xwalk.csv.gz'.format(state=state)

        state_url = 'https://lehd.ces.census.gov/data/lodes/LODES7/' + state
        od_url = state_url + '/od/' + od_fname
        self.od_fname = self.cache_file(od_url)

        xwalk_url = state_url + '/' + xwalk_fname
        self.xwalk_fname = self.cache_file(xwalk_url)



from typing import Optional

import pandas as pd
import pycombo

from sttn.network import SpatioTemporalNetwork


def combo_communities(data: SpatioTemporalNetwork, weight: Optional[str] = None, **kwargs) -> SpatioTemporalNetwork:
    graph = data.to_multigraph()
    partition_dict, _ = pycombo.execute(graph, weight=weight, **kwargs)
    partition_df = pd.DataFrame.from_dict(partition_dict, orient='index', columns=['cluster'])
    data_with_community = data.join_node_labels(partition_df)
    return data_with_community



import os

import geopandas as gpd
import numpy as np
import pandas as pd
import pyarrow.dataset as ds

from sttn import constants
from sttn import network
from typing import Optional
from .data_provider import DataProvider

PRAGUE_DISTRICT_CODES = ['CZ0101', 'CZ0102', 'CZ0103', 'CZ0104', 'CZ0105', 'CZ0106',
                         'CZ0107', 'CZ0108', 'CZ0109', 'CZ010A']
PRAGUE_CODE = 'CZ0100'
HEALTHCARE_DATA_VAR = "STTN_CZ_HEALTH_DATA"


class JourneyDataProvider(DataProvider):
    """Brno journey data provider. Builds a network where every node represents a city district
    and every edge contains the number of people who commuted from the origin to the destination
    district within a given hour. The dataset covers one week of data from 7th to 13th October 2019.
    The data is based on the passive communication between mobile phones and cell towers,
    and it covers only one mobile phone company.
    """

    @staticmethod
    def read_trip_file(file_path: str, start_level: int, end_level: int) -> pd.DataFrame:
        trips = pd.read_csv(file_path)
        # keep only trips with known origin and destination
        trips_filtered = trips[~trips['start_kod'].isnull() & ~trips['cil_kod'].isnull()]
        # filter Brno trip levels (urban/suburban)
        trips_filtered = trips_filtered[
            (trips_filtered['start_level'] == start_level) & (trips_filtered['cil_level'] == end_level)]
        trips_filtered = trips_filtered.astype({"start_kod": int, "cil_kod": int})
        columns_to_rename = {"start_cas": "start_hr", "cil_cas": "end_hr", "start_kod": constants.ORIGIN,
                             "cil_kod": constants.DESTINATION, "pocet": "count", "cz": "is_czech",
                             "pocet_kalibrovano": "count_adjusted"}
        renamed = trips_filtered.rename(columns=columns_to_rename)

        if len(renamed.day.value_counts()) != 1:
            raise ValueError(f"The day file contains data from multiple days: {renamed.day_.value_counts()}")

        day_delta = renamed.day.iloc[0]
        # observation start date is 7.10.2019
        renamed['start_ts'] = renamed.start_hr.apply(
            lambda hr: pd.Timestamp(year=2019, month=10, day=(7 + day_delta), hour=hr))
        renamed['end_ts'] = renamed.end_hr.apply(
            lambda hr: pd.Timestamp(year=2019, month=10, day=(7 + day_delta), hour=hr))
        columns_to_keep = [constants.ORIGIN, constants.DESTINATION, "is_czech", "count", "count_adjusted", "start_ts",
                           "end_ts"]
        return renamed[columns_to_keep]

    @staticmethod
    def read_edges(journey_csv_folder: str, start_level: int, end_level: int) -> pd.DataFrame:
        edge_files = os.listdir(journey_csv_folder)
        edge_dfs = [JourneyDataProvider.read_trip_file(f"{journey_csv_folder}/{fname}", start_level, end_level) for
                    fname in edge_files if
                    fname.endswith(".csv")]
        return pd.concat(edge_dfs)

    @staticmethod
    def read_nodes(shapefile_file_name: str) -> gpd.GeoDataFrame:
        nodes = gpd.read_file(shapefile_file_name)
        columns_to_keep = ["KOD_KU", "NAZ_KU", "KOD_ZUJ", "NAZ_ZUJ", "geometry"]
        nodes = nodes[columns_to_keep]
        column_names = {"KOD_KU": constants.NODE_ID, "NAZ_KU": "block", "KOD_ZUJ": "district_code",
                        "NAZ_ZUJ": "district"}
        renamed = nodes.rename(columns=column_names)
        types = {constants.NODE_ID: int, 'district_code': int}
        converted = renamed.astype(types).set_index(constants.NODE_ID).to_crs(epsg=4326)
        return converted

    @staticmethod
    def get_data(journey_csv_folder: str, shapefile_name: str, start_level: int = 1,
                 end_level: int = 1) -> network.SpatioTemporalNetwork:
        nodes = JourneyDataProvider.read_nodes(shapefile_name)
        edges = JourneyDataProvider.read_edges(journey_csv_folder, start_level=start_level, end_level=end_level)
        sttn_network = network.SpatioTemporalNetwork(nodes=nodes, edges=edges)
        return sttn_network


class HealthcareDataProvider(DataProvider):
    """
    Czech Republic healthcare data provider. Builds a network where every node represents a Czech Republic district
    and every edge contains aggregated patients information where the origin is the residence of patients, and 
    the destination is a healthcare facility district they commuted to. The provider gives information on how many visits
    happened from origin to destination districts in a specific month of the year for different specializations and ICD-10
    disease categories. The dataset covers 13 years of monthly data from 2010-01 to 2022-12. 
    """
    @staticmethod
    def get_data(year: Optional[str], data_folder: Optional[str] = None) -> network.SpatioTemporalNetwork:
        """
        Retrieves Czech Republic Healthcare data. 
        Args:
            year (year): 4-digit year of healthcare data

        Returns:
            SpatioTemporalNetwork: An STTN network where nodes represent the Czech Republic districts 
                (also called 'okres'), and edges represent the patients' treatment mobility and details.

            The nodes dataframe contains the following columns:
                'id' (str) - district LAU code, Pandas index column
                'name' (str) - district names
                'geometry' (shape) - shape object for the district

            The edges dataframe contains the following columns:
                'origin' (str) - patients' residence district id
                'destination' (str) - patients' medical facility district id
                'year_visit' (int) - patients' year of visit
                'month_visit' (int) - patients' month of visit
                'specialization' (int) - patients' treatment medical branch specialization number
                'icd10_category' (int) -patients' disease ICD10 category number
                'number_of_visits' (int) - total number of patients who visited from a residence district to a healthcare facility,
                                           grouped by medical specialization and disease category

        
        """
        folder_env = os.getenv(HEALTHCARE_DATA_VAR)
        folder = data_folder or folder_env
        if folder is None:
            raise ValueError(f"data_folder provider argument and {HEALTHCARE_DATA_VAR} variable are not set")
      
        nodes = HealthcareDataProvider.read_nodes(f"{folder}/lau1.geojson")
        edges = HealthcareDataProvider.read_edges(year=int(year), data_file=f"{folder}/healthcare.parquet")

        ids_to_keep = nodes.index
        edges = edges[
            edges.origin.isin(ids_to_keep) & edges.destination.isin(ids_to_keep)]

        nodes_to_keep = np.union1d(edges.origin.unique(), edges.destination.unique())
        nodes = nodes[nodes.index.isin(nodes_to_keep)]
        sttn_network = network.SpatioTemporalNetwork(nodes=nodes, edges=edges)
        return sttn_network

    @staticmethod
    def read_edges(year: Optional[int], data_file: str) -> pd.DataFrame:
        dataset = ds.dataset(data_file, format='parquet')
        if not year is None:
            table = dataset.to_table(filter=ds.field('year_visit') == year)
        else:
            table = dataset.to_table()

        edges = table.to_pandas()
        columns_to_rename = {"okres_residence": constants.ORIGIN,
                             "okres_servise": constants.DESTINATION, }
        renamed = edges.rename(columns=columns_to_rename)
        # map Prague district codes to LAU1
        prague_dict = {key: PRAGUE_CODE for key in PRAGUE_DISTRICT_CODES}
        renamed[constants.ORIGIN] = renamed[constants.ORIGIN].replace(prague_dict)
        renamed[constants.DESTINATION] = renamed[constants.DESTINATION].replace(prague_dict)
        return renamed

    @staticmethod
    def read_nodes(geojson_file: str) -> gpd.GeoDataFrame:
        nodes = gpd.read_file(geojson_file)
        columns_to_keep = ["lau", "name", "geometry"]
        nodes = nodes[columns_to_keep]
        column_names = {"lau": constants.NODE_ID}
        renamed = nodes.rename(columns=column_names)
        converted = renamed.set_index(constants.NODE_ID).to_crs(epsg=4326)
        return converted



import geopandas as gpd
import pandas as pd

from sttn.network import SpatioTemporalNetwork


def read_parquet(path: str) -> SpatioTemporalNetwork:
    """Read STTN nodes and edges to the Parquet format.
    """
    node_path = f"{path}-nodes.parquet"
    edge_path = f"{path}-edges.parquet"
    nodes = gpd.read_parquet(node_path)
    edges = pd.read_parquet(edge_path)
    return SpatioTemporalNetwork(nodes=nodes, edges=edges)
